/**
  @file dbus.c

  @author Johan Hedberg <johan.hedberg@nokia.com>

  Copyright (C) 2004 Nokia. All rights reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

*/
#include <stdlib.h>
#include <stdio.h>

#include <glib.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>

#include <bt-dbus.h>

#include "dbus.h"

static char *used_dev = NULL;

static char *get_rfcomm_dev(DBusConnection *c, const char *svc) {
    char *rfcomm_dev = NULL, *dev;
    gboolean ret;
    DBusError err;
    DBusMessage *msg, *reply;

    msg = dbus_message_new_method_call(BTCOND_SERVICE,
                                       BTCOND_REQ_PATH,
                                       BTCOND_REQ_INTERFACE,
                                       BTCOND_RFCOMM_CONNECT_REQ);
    if (msg == NULL) {
        fprintf(stderr, "Out of memory during dbus_message_new_method_call()\n");
        exit(EXIT_FAILURE);
    }

    if (!dbus_message_append_args(msg,
                                  DBUS_TYPE_STRING, svc,
                                  DBUS_TYPE_BOOLEAN, TRUE,
                                  DBUS_TYPE_INVALID)) {
        fprintf(stderr, "Out of memory!\n");
        exit(EXIT_FAILURE);
    }

    dbus_error_init(&err);
    reply = dbus_connection_send_with_reply_and_block(c, msg, 1000 * 20, &err);
    dbus_message_unref(msg);
    if (dbus_error_is_set(&err)) {
        printf("btcond returned error: %s\n", err.name);
        dbus_error_free(&err);
        return NULL;
    }

    ret = dbus_message_get_args(reply, NULL,
                                DBUS_TYPE_STRING, &rfcomm_dev,
                                DBUS_TYPE_INVALID);
    dbus_message_unref(reply);
    if (ret == FALSE) {
        printf("get_args failed\n");
        return NULL;
    }

    dev = g_strdup(rfcomm_dev);
    g_free(rfcomm_dev);

    return dev;
}

char *get_obex_dev(gboolean ftp) {
    char *rfcomm_dev = NULL;
    DBusConnection *c;

    c = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
    if (c == NULL) {
        fprintf(stderr, "Unable to connect to system D-BUS\n");
        return NULL;
    }

    if (ftp) {
        rfcomm_dev = get_rfcomm_dev(c, "NFTP"); 
        if (rfcomm_dev == NULL) {
            rfcomm_dev = get_rfcomm_dev(c, "FTP"); 
        }
    }
    else {
        rfcomm_dev = get_rfcomm_dev(c, "OPP"); 
    }

    if (rfcomm_dev != NULL) {
        g_free(used_dev);
        used_dev = g_strdup(rfcomm_dev);
    }

    return rfcomm_dev;
}

void disconnect_obex_dev(void) {
    DBusError err;
    DBusConnection *c;
    DBusMessage *msg, *reply;

    if (used_dev == NULL) {
        return;
    }

    c = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
    if (c == NULL) {
        fprintf(stderr, "Unable to connect to system D-BUS\n");
        return;
    }

    msg = dbus_message_new_method_call(BTCOND_SERVICE,
                                       BTCOND_REQ_PATH,
                                       BTCOND_REQ_INTERFACE,
                                       BTCOND_RFCOMM_DISCONNECT_REQ);
    if (msg == NULL) {
        fprintf(stderr, "Out of memory during dbus_message_new_method_call()\n");
        exit(EXIT_FAILURE);
    }

    if (!dbus_message_append_args(msg,
                                  DBUS_TYPE_STRING, used_dev,
                                  DBUS_TYPE_INVALID)) {
        fprintf(stderr, "Out of memory!\n");
        exit(EXIT_FAILURE);
    }

    dbus_error_init(&err);
    reply = dbus_connection_send_with_reply_and_block(c, msg, 1000 * 20, &err);
    dbus_message_unref(msg);
    if (dbus_error_is_set(&err)) {
        printf("btcond returned error: %s\n", err.name);
        dbus_error_free(&err);
    }

    g_free(used_dev);
    used_dev = NULL;
}

